#ifndef __DMATOSP_H__
#define __DMATOSP_H__


#include "dmawrap.h"


class CDmaToSP : public CDmaWrap
{
public:
	CDmaToSP(void * pDmaMem, uint32 uDmaQuadSize, TRANS_MODE trans_mode=TRANS_CHAINED) : CDmaWrap(pDmaMem,uDmaQuadSize,trans_mode) {m_bPrepDone=false;}

	void PrepareLaunch(uint32 uToSADR);	// specific for SP channels
	void Launch(int iSPS2Device=-1);
	void Wait();

private:
	bool m_bPrepDone;	// false if launch preperations haven't been made
	uint32 m_uCHCRVal;	// generated value for the CHCR register
};


inline void CDmaToSP::PrepareLaunch(uint32 uToSADR)
{
	asm("sync.l");
	assert(((*EE_D9_CHCR)&0x0100)==0 && m_bPrepDone==false);	// Can't launch if already running

	switch(m_uTransMode)
	{
		case TRANS_NORMAL:
			*EE_D9_MADR = (uint32) m_pStartPhys&0x0fffffff;
			*EE_D9_SADR = (uint32) uToSADR&0x3fff;
			*EE_D9_QWC = (uint32) m_uDmaQuadSize;
			m_uCHCRVal = GenCHCR(0,0,0,0,0,1);
			break;
		case TRANS_CHAINED:
			*EE_D9_QWC = 0;
			*EE_D9_TADR = (uint32) m_pStartPhys&0x0fffffff;
			*EE_D9_SADR = (uint32) uToSADR&0x3fff;
			m_uCHCRVal = GenCHCR(0,1,0,0,0,1);
			break;
		default:
			assert(false);	// unknown specification
	}

	assert((m_bPrepDone=true) == true);			// I agree it's repulsive but it works this way
}


inline void CDmaToSP::Launch(int iSPS2Device)
{
	assert(m_bPrepDone == true);

#ifdef DEBUG_DMA
	if (iSPS2Device>=0) {
	  sps2ValidateDMA(iSPS2Device,9,m_uCHCRVal,0);
	}
#endif

	*EE_D9_CHCR = m_uCHCRVal;

	assert((m_bPrepDone=false) == false);			// I agree it's repulsive but it works this way
}

inline void CDmaToSP::Wait()
{
	while((*EE_D9_CHCR)&0x0100)
		;
}






#endif
