#ifndef __DMAWRAP_H__
#define __DMAWRAP_H__


#include <stdio.h>
#include <string.h>
#include "p2lassert.h"
#include "exttypes.h"
#include "macro_codes.h"
#include "sps2lib.h"


#define SANITYCHECK(uAddTo)		assert( m_pDmaMem!=NULL && m_uDmaQuadSize>0 && ((uint32(m_pCurLocation)&(uAddTo-1))==0) && (((uint32(m_pCurLocation)+uAddTo) - uint32(m_pDmaMem)) <= (m_uDmaQuadSize<<4))    )
#define DMAFROMSP				0x80000000

class CDmaWrap
{
	friend void  DistribGlobPtrs(void * pDescCached, void * pDescUncached);

public:	// constructers and destructers
	enum TRANS_MODE
	{
		TRANS_NORMAL=1,
		TRANS_CHAINED
	};

	void Reset();
	CDmaWrap(void * pDmaMem, uint32 uDmaQuadSize, TRANS_MODE trans_mode=TRANS_CHAINED);
	~CDmaWrap();

public:
	// Source Tags
	void AttachSrcRefTag(void * pData, uint32 uQuadSize, uint64 uTTECode=0);
	void AttachSrcCntTag(uint64 uTTECode=0);
	void AttachSrcEndTag(uint64 uTTECode=0);
	void AttachSrcNextTag(void * pData, uint64 uTTECode=0);
	void AttachSrcRefsTag(void * pData, uint32 uQuadSize, uint64 uTTECode=0);
	void AttachSrcRefeTag(void * pData, uint32 uQuadSize, uint64 uTTECode=0);
	void AttachSrcCallTag(void * pData, uint64 uTTECode=0);
	void AttachSrcRetTag(uint64 uTTECode=0);

	
	// Destination Tags
	void AttachDstCntTag(void * pData, uint64 uTTECode=0);
	void AttachDstCntsTag(void * pData, uint64 uTTECode=0);
	void AttachDstEndTag(void * pData, uint64 uTTECode=0);

	// To close a packet
	void EndPacket();


public:	// Data section
	void add8(uint8 uVal);
	void add16(uint16 uVal);
	void add32(uint32 uVal);
	void add64(uint64 uVal);
	void add128(uint128 uVal);
	void addData(void * pData, uint32 uByteSize);

public:	// pointer tweaks
	void * GetCurLocation();
	void SetCurLocation(void * pCurLoc);
	void DisableUCAB();
	void EnableUCAB();	

protected:
	uint32 GenCHCR(uint32 uDir, uint32 uMod, uint32 uAsp, uint32 uTTE, uint32 uTIE, uint32 uStr);

	void * m_pDmaMem;
	uint32 m_uDmaQuadSize;	// Buffer size
	TRANS_MODE m_uTransMode;
	bool m_bSetTTE;

protected:
	// PS2Linux specific details
	void * m_pStartPhys;	// physical value of m_pDmaMem

private:
	// PS2Linux specific details
	void * VirtToPhys(void * pVirtPtr);
	void * ForceToInternalPtr(void * pPtrIn);
	void * ForceToCachedVirtPtr(void * pPtrIn);
	void * ForceToUncachedVirtPtr(void * pPtrIn);
	
	static sps2Memory_t * m_pDescCached;
	static sps2Memory_t * m_pDescUncached;

	bool m_bTransferFromScratchPad;
	bool m_bCacheEnabled;

private:
	void AddTag(uint64 uTag, uint64 uTTECode);

	void * m_pCurLocation;

private:	// Packet logic
	void AddTagAndPacket(uint64 uTag, uint64 uTTECode);

	bool m_bPacketInDevelopment;
	void * m_pRecordLocation;	// used to update TagQWCnt when packet done
};


// Add DMA-Tags
inline void CDmaWrap::AttachSrcRefTag(void * pData, uint32 uQuadSize, uint64 uTTECode)
{
	AddTag(GEN_REF_TAG(VirtToPhys(pData), uQuadSize), uTTECode);
}	
	
inline void CDmaWrap::AttachSrcCntTag(uint64 uTTECode)
{
	AddTagAndPacket(GEN_CNT_TAG(0), uTTECode);
}

inline void CDmaWrap::AttachSrcEndTag(uint64 uTTECode)
{
	AddTagAndPacket(GEN_END_TAG(0), uTTECode);
}

inline void CDmaWrap::AttachSrcNextTag(void * pData, uint64 uTTECode)
{
	AddTagAndPacket(GEN_NEXT_TAG(VirtToPhys(pData), 0), uTTECode);
}

inline void CDmaWrap::AttachSrcRefsTag(void * pData, uint32 uQuadSize, uint64 uTTECode)
{
	AddTag(GEN_REFS_TAG(VirtToPhys(pData), uQuadSize), uTTECode);
}

inline void CDmaWrap::AttachSrcRefeTag(void * pData, uint32 uQuadSize, uint64 uTTECode)
{
	AddTag(GEN_REFE_TAG(VirtToPhys(pData), uQuadSize), uTTECode);
}

inline void CDmaWrap::AttachSrcCallTag(void * pData, uint64 uTTECode)
{
	AddTagAndPacket(GEN_CALL_TAG(VirtToPhys(pData), 0), uTTECode);
}

inline void CDmaWrap::AttachSrcRetTag(uint64 uTTECode)
{
	AddTagAndPacket(GEN_RET_TAG(0), uTTECode);
}

inline void CDmaWrap::AttachDstCntTag(void * pData, uint64 uTTECode)
{
	AddTagAndPacket(GEN_CNT_DSTTAG(VirtToPhys(pData), 0), uTTECode);
}

inline void CDmaWrap::AttachDstCntsTag(void * pData, uint64 uTTECode)
{
	AddTagAndPacket(GEN_CNTS_DSTTAG(VirtToPhys(pData), 0), uTTECode);
}

inline void CDmaWrap::AttachDstEndTag(void * pData, uint64 uTTECode)
{
	AddTagAndPacket(GEN_END_DSTTAG(VirtToPhys(pData), 0), uTTECode);
}

inline void CDmaWrap::AddTag(uint64 uTag, uint64 uTTECode)
{
	SANITYCHECK(16);
	assert(uTTECode==0 || m_bSetTTE==true);
	assert(!m_bPacketInDevelopment);
	*((uint64 *) m_pCurLocation)++ = uTag;
	*((uint64 *) m_pCurLocation)++ = uTTECode;
}

inline void CDmaWrap::AddTagAndPacket(uint64 uTag, uint64 uTTECode)
{	
	SANITYCHECK(16);
	assert(uTTECode==0 || m_bSetTTE==true);
	assert(!m_bPacketInDevelopment);
	m_pRecordLocation = m_pCurLocation;
	*((uint64 *) m_pCurLocation)++ = uTag;
	*((uint64 *) m_pCurLocation)++ = uTTECode;
	assert((m_bPacketInDevelopment=true) == true);	// This will be erased in release build
}

inline void CDmaWrap::EndPacket()
{
	assert(m_bPacketInDevelopment);
	assert(((uint32) m_pCurLocation) > ((uint32) m_pRecordLocation));
	uint32 uOffs = ((uint32) m_pCurLocation) - ((uint32) m_pRecordLocation);
	assert((uOffs&0xf)==0);	// CurrentLocation must be Quad aligned to endPacket
	uint32 uSize = (uOffs>>4)-1;	// -1 to remove offs of the Tag
	((uint64 *)m_pRecordLocation)[0] |= uSize;	// insert QuadCount
	assert((m_bPacketInDevelopment=false) == false);	// This will be erased in release build
}


// Add Data functions
inline void CDmaWrap::add8(uint8 uVal)
{
	SANITYCHECK(1);
	assert(m_bPacketInDevelopment || m_uTransMode==TRANS_NORMAL);
	*((uint8 *) m_pCurLocation)++ = uVal;
}

inline void CDmaWrap::add16(uint16 uVal)
{
	SANITYCHECK(2);
	assert(m_bPacketInDevelopment || m_uTransMode==TRANS_NORMAL);
	*((uint16 *) m_pCurLocation)++ = uVal;
}

inline void CDmaWrap::add32(uint32 uVal)
{
	SANITYCHECK(4);
	assert(m_bPacketInDevelopment || m_uTransMode==TRANS_NORMAL);
	*((uint32 *) m_pCurLocation)++ = uVal;
}

inline void CDmaWrap::add64(uint64 uVal)
{
	SANITYCHECK(8);
	assert(m_bPacketInDevelopment || m_uTransMode==TRANS_NORMAL);
	*((uint64 *) m_pCurLocation)++ = uVal;
}

inline void CDmaWrap::add128(uint128 uVal)
{
	SANITYCHECK(16);
	assert(m_bPacketInDevelopment || m_uTransMode==TRANS_NORMAL);
	*((uint128 *) m_pCurLocation)++ = uVal;
}

inline void CDmaWrap::addData(void * pData, uint32 uByteSize)
{
	assert( m_pDmaMem!=NULL && m_uDmaQuadSize>0 && (((uint32(m_pCurLocation)+uByteSize) - uint32(m_pDmaMem)) <= (m_uDmaQuadSize<<4))    );

	assert(m_bPacketInDevelopment || m_uTransMode==TRANS_NORMAL);
	memcpy(m_pCurLocation, pData, uByteSize);
	((char *) m_pCurLocation) += uByteSize;
}

// Reset Chain
inline void CDmaWrap::Reset()
{
	m_pCurLocation = m_pDmaMem;
}

// Get/Set current locaton
inline void * CDmaWrap::GetCurLocation() 
{ 
	return m_pCurLocation;
}

inline void CDmaWrap::SetCurLocation(void * pCurLoc)
{
	m_pCurLocation=ForceToInternalPtr(pCurLoc);
	assert(uint32(m_pDmaMem)<=uint32(m_pCurLocation));
	assert(uint32(m_pCurLocation)<uint32(m_pDmaMem)+(m_uDmaQuadSize<<4));
}


// CHCR builder
inline uint32 CDmaWrap::GenCHCR(uint32 uDir, uint32 uMod, uint32 uAsp, uint32 uTTE, uint32 uTIE, uint32 uStr)
{
	return	(uDir<<0) |		// direction from mem (effective only on Vif1)
			(uMod<<2) |     // Mode: Chain
			(uAsp<<4) |		// Address stack ptr: no address pushed by call tag
			(uTTE<<6) |		// Tag Transfer Enable: Yes
			(uTIE<<7) |		// Tag Interrupt Enable: Disables IRQ bit of DMAtag
			(uStr<<8);		// Start: start dma
}



#endif
