#include "cboxmesh.h"
#include <stdio.h>
#include "p2lassert.h"
#include <stdlib.h>
#include <math.h>


bool CBoxMesh::GenerateBox(CDmaVif1 * pDmaVif1, int iWidthSegs, float fWidthDim, int iHeightSegs, float fHeightDim, 
						   int iLengthSegs, float fLengthDim)
{
	CleanUp();
	pDmaVif1->Reset();
	m_pDmaTags = pDmaVif1->GetCurLocation();

	m_iWidthSegs = iWidthSegs;
	m_iHeightSegs = iHeightSegs;
	m_iLengthSegs = iLengthSegs; 

	m_fWidthDim = fWidthDim;
	m_fHeightDim = fHeightDim;
	m_fLengthDim = fLengthDim;

	// Interior vertices
	// 0, 3 Xplane
	m_iNumOfUniqueVerts += (2*(iLengthSegs-1)*(iHeightSegs-1));
	

	// 1, 4 Yplane
	m_iNumOfUniqueVerts += (2*(iWidthSegs-1)*(iLengthSegs-1));


	// 2, 5 Zplane
	m_iNumOfUniqueVerts += (2*(iWidthSegs-1)*(iHeightSegs-1));

	// Edge Vertices
	m_iNumOfUniqueVerts += (4*(iLengthSegs-1));
	m_iNumOfUniqueVerts += (4*(iWidthSegs-1));
	m_iNumOfUniqueVerts += (4*(iHeightSegs-1));

	// Corner vertices
	m_iNumOfUniqueVerts += 8;

	m_pUniCols = NULL;
	m_pUniVerts = NULL;
	m_iPointID = NULL;

	if((m_pUniCols=new int[m_iNumOfUniqueVerts*4]) == NULL)
	{ CleanUp(); return false; } 
	else if((m_pUniVerts=new float[m_iNumOfUniqueVerts*4]) == NULL)
	{ CleanUp(); return false; } 
	else if((m_iPointID=new int[m_iNumOfUniqueVerts]) == NULL)
	{ CleanUp(); return false; }


	for(int i=0; i<m_iNumOfUniqueVerts; i++)
		m_iPointID[i] = -1;		// No point found

	// Alloc total amount of strips
	SRenderGadget * pAllStrips = NULL; 
	int iTotalNumOfStrips = 0;

	// 0, 3 Xplane
	iTotalNumOfStrips += (2*iHeightSegs);
	

	// 1, 4 Yplane
	iTotalNumOfStrips += (2*iLengthSegs);


	// 2, 5 Zplane
	iTotalNumOfStrips += (2*iHeightSegs);


	if((pAllStrips=new SRenderGadget[iTotalNumOfStrips]) == NULL)
	{ CleanUp(); return false; }

	for(int iStrip=0; iStrip<iTotalNumOfStrips; iStrip++)
	{
		pAllStrips[iStrip].m_pVerts = NULL;
		pAllStrips[iStrip].m_pCols = NULL;
		pAllStrips[iStrip].m_pTexCrds = NULL;
	}

	SRenderGadget * pCurStrips = pAllStrips;

	if(!GenerateSurf(0, iLengthSegs, iHeightSegs, pCurStrips))
	{ CleanStrips(pAllStrips, iTotalNumOfStrips); CleanUp(); return false; }
	pCurStrips += iHeightSegs;
	if(!GenerateSurf(1, iWidthSegs, iLengthSegs, pCurStrips))
	{ CleanStrips(pAllStrips, iTotalNumOfStrips); CleanUp(); return false; }
	pCurStrips += iLengthSegs;
	if(!GenerateSurf(2, iWidthSegs, iHeightSegs, pCurStrips))
	{ CleanStrips(pAllStrips, iTotalNumOfStrips); CleanUp(); return false; }
	pCurStrips += iHeightSegs;
	if(!GenerateSurf(3, iLengthSegs, iHeightSegs, pCurStrips))
	{ CleanStrips(pAllStrips, iTotalNumOfStrips); CleanUp(); return false; }
	pCurStrips += iHeightSegs;
	if(!GenerateSurf(4, iWidthSegs, iLengthSegs, pCurStrips))
	{ CleanStrips(pAllStrips, iTotalNumOfStrips); CleanUp(); return false; }
	pCurStrips += iLengthSegs;
	if(!GenerateSurf(5, iWidthSegs, iHeightSegs, pCurStrips))
	{ CleanStrips(pAllStrips, iTotalNumOfStrips); CleanUp(); return false; }
		
	bool bRes = ConvertToChunk(pDmaVif1, pAllStrips, iTotalNumOfStrips);
	CleanStrips(pAllStrips, iTotalNumOfStrips);

	return bRes;
}

#ifndef max
	#define max(a,b) ((a>b) ? a : b )
#endif

#ifndef min
	#define min(a,b) ((a<b) ? a : b )
#endif

bool CBoxMesh::GenerateSurf(int iSurf, int iWDim, int iHDim, SRenderGadget * pCurStrips)
{
	int iCurStrip = 0;

	for(int j=0; j<iHDim; j++)
	{
		pCurStrips[iCurStrip].m_iNumElems = 2*(iWDim+1);
		pCurStrips[iCurStrip].m_iPrimType = PRIMTYPE_STRIP;
		if((pCurStrips[iCurStrip].m_pVerts=new float[4*pCurStrips[iCurStrip].m_iNumElems]) == NULL)
			return false;
		else if((pCurStrips[iCurStrip].m_pCols=new int[4*pCurStrips[iCurStrip].m_iNumElems]) == NULL)
			return false;
		else if((pCurStrips[iCurStrip].m_pTexCrds=new float[4*pCurStrips[iCurStrip].m_iNumElems]) == NULL)
			return false;

		for(int i=0; i<=iWDim; i++)
		{
			GetPoint(iSurf, i, j, &pCurStrips[iCurStrip].m_pVerts[4*(i<<1)], &pCurStrips[iCurStrip].m_pCols[4*(i<<1)]);
			GetPoint(iSurf, i, j+1, &pCurStrips[iCurStrip].m_pVerts[4*((i<<1)+1)], &pCurStrips[iCurStrip].m_pCols[4*((i<<1)+1)]);
			
			float fScale = 1.0f;
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+0) + 0] = ((float(i)/iWDim)*fScale);
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+0) + 1] = ((float(j)/iHDim)*fScale);
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+0) + 2] = 0;
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+0) + 3] = 0;

			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+1) + 0] = ((float(i)/iWDim)*fScale);
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+1) + 1] = ((float(j+1)/iHDim)*fScale);
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+1) + 2] = 0;
			pCurStrips[iCurStrip].m_pTexCrds[4*((i<<1)+1) + 3] = 0;
		}

		++iCurStrip;
	}

	// Build Normals
	iCurStrip = 0;

	float fBoxNorm[3];

	fBoxNorm[0]=1; fBoxNorm[1]=0; fBoxNorm[2]=0;

	if((iSurf%3)==1)
	{ fBoxNorm[0]=0; fBoxNorm[1]=1; }
	else if((iSurf%3)==2)
	{ fBoxNorm[0]=0; fBoxNorm[2]=1; }

	if(iSurf > 2)
	{ fBoxNorm[0] = -fBoxNorm[0]; fBoxNorm[1] = -fBoxNorm[1]; fBoxNorm[2] = -fBoxNorm[2]; }


	for(int j=0; j<iHDim; j++)
	{
		SRenderGadget &sStrip = pCurStrips[iCurStrip];


		for(int i=0; i<=iWDim; i++)
		{
			float fNorm[4];
			
			GetNormal(iSurf, i, j, iWDim, iHDim, fNorm);

			if((fBoxNorm[0]*fNorm[0] + fBoxNorm[1]*fNorm[1] + fBoxNorm[2]*fNorm[2])<0)
			{
				fNorm[0] = -fNorm[0]; fNorm[1] = -fNorm[1]; fNorm[2] = -fNorm[2]; 
			}


			/*int uVal1 = (fNorm[0]+1)*0.5f*4095;
			int uVal2 = (fNorm[1]+1)*0.5f*4095;
			uVal1 = min(0xfff, uVal1);
			uVal1 = max(0x0, uVal1);
			uVal2 = min(0xfff, uVal2);
			uVal2 = max(0x0, uVal2);
			*((int *) &sStrip.m_pTexCrds[4*((i<<1)+0) + 3]) = ((uVal2<<12) | uVal1);*/

			sStrip.m_pTexCrds[4*((i<<1)+0) + 2] = fNorm[0];
			sStrip.m_pTexCrds[4*((i<<1)+0) + 3] = fNorm[1];
			sStrip.m_pVerts[4*((i<<1)+0) + 3] = fNorm[2];

			GetNormal(iSurf, i, j+1, iWDim, iHDim, fNorm);

			if((fBoxNorm[0]*fNorm[0] + fBoxNorm[1]*fNorm[1] + fBoxNorm[2]*fNorm[2])<0)
			{
				fNorm[0] = -fNorm[0]; fNorm[1] = -fNorm[1]; fNorm[2] = -fNorm[2]; 
			}

			/*uVal1 = (fNorm[0]+1)*0.5f*4095;
			uVal2 = (fNorm[1]+1)*0.5f*4095;
			uVal1 = min(0xfff, uVal1);
			uVal1 = max(0x0, uVal1);
			uVal2 = min(0xfff, uVal2);
			uVal2 = max(0x0, uVal2);
		
			*((int *) &sStrip.m_pTexCrds[4*((i<<1)+1) + 3]) = ((uVal2<<12) | uVal1);*/
			sStrip.m_pTexCrds[4*((i<<1)+1) + 2] = fNorm[0];
			sStrip.m_pTexCrds[4*((i<<1)+1) + 3] = fNorm[1];
			sStrip.m_pVerts[4*((i<<1)+1) + 3] = fNorm[2];
		}

		++iCurStrip;
	}

	return true;
}



float fRnd() { return ((float) (((unsigned int) rand()) & 0xffff)) / ((unsigned int) 0xffff); }

void AddToSum(float *fNormSum, float *fP1, float *fP2, float *fP3)
{
	float v1[] = {fP2[0]-fP1[0], fP2[1]-fP1[1], fP2[2]-fP1[2]};
	float v2[] = {fP3[0]-fP1[0], fP3[1]-fP1[1], fP3[2]-fP1[2]};
	float N[3];


	N[0] = v1[1]*v2[2] - v1[2]*v2[1];
	N[1] = v1[2]*v2[0] - v1[0]*v2[2];
	N[2] = v1[0]*v2[1] - v1[1]*v2[0];

	for(int i=0; i<3; i++)
		fNormSum[i] += N[i];
}

void CBoxMesh::GetNormal(int iSurf, int i, int j, int iWDim, int iHDim, float * pNorm)
{
	float fNormSum[] = {0, 0, 0, 0};
	float fP1[4], fP2[4], fP3[4];
	int pCol[4];

	GetPoint(iSurf,i, j, fP1, pCol);

	if(i-1>=0)
	{
		if(j-1>=0)
		{
			GetPoint(iSurf,i, j-1, fP2, pCol);
			GetPoint(iSurf,i-1, j-1, fP3, pCol);
			AddToSum(fNormSum, fP1, fP2, fP3);
			GetPoint(iSurf,i-1, j, fP2, pCol);
			AddToSum(fNormSum, fP1, fP3, fP2);
		}
		
		if(j+1<=iHDim)
		{
			GetPoint(iSurf,i-1, j, fP2, pCol);
			GetPoint(iSurf,i, j+1, fP3, pCol);
			AddToSum(fNormSum, fP1, fP2, fP3);
		}
	}
	if(i+1<=iWDim)
	{
		if(j-1>=0)
		{
			GetPoint(iSurf,i+1, j, fP2, pCol);
			GetPoint(iSurf,i, j-1, fP3, pCol);
			AddToSum(fNormSum, fP1, fP2, fP3);
		}
		
		if(j+1<=iHDim)
		{
			GetPoint(iSurf,i, j+1, fP2, pCol);
			GetPoint(iSurf,i+1, j+1, fP3, pCol);
			AddToSum(fNormSum, fP1, fP2, fP3);
			GetPoint(iSurf,i+1, j, fP2, pCol);
			AddToSum(fNormSum, fP1, fP3, fP2);
		}
	}

	float fInvLength = sqrt(fNormSum[0]*fNormSum[0] + fNormSum[1]*fNormSum[1] + fNormSum[2]*fNormSum[2]);
	if(fInvLength != 0) fInvLength = 1 / fInvLength;

	for(int i=0; i<3; i++)
	{
		pNorm[i] = fInvLength * fNormSum[i];
		if(pNorm[i] < -1) pNorm[i]=-1;
		else if(pNorm[i] > 1) pNorm[i]=1;
	}

	pNorm[3] = 0;
}

void CBoxMesh::GetPoint(int iSurf, int i, int j, float * pVert, int * pCol)
{
	int iXWSeg, iYHSeg;
	float fXWDim, fYHDim, fZDDim;
	int iID;

	switch(iSurf%3)	
	{
		case	0:		// Xplane
			iXWSeg = m_iLengthSegs;
			iYHSeg = m_iHeightSegs;
			fXWDim = m_fLengthDim;
			fYHDim = m_fHeightDim;
			fZDDim = m_fWidthDim;
			iID = (m_iHeightSegs+1)*(m_iWidthSegs+1)*i + (m_iWidthSegs+1)*j;
			if(iSurf<3) iID += m_iWidthSegs;
			break;
		case	1:		// Yplane
			iXWSeg = m_iWidthSegs;
			iYHSeg = m_iLengthSegs;
			fXWDim = m_fWidthDim;
			fYHDim = m_fLengthDim;
			fZDDim = m_fHeightDim;
			iID = (m_iHeightSegs+1)*(m_iWidthSegs+1)*j + i;
			if(iSurf<3) iID += ((m_iWidthSegs+1)*m_iHeightSegs);
			break;
		case	2:		// Zplane
			iXWSeg = m_iWidthSegs;
			iYHSeg = m_iHeightSegs;
			fXWDim = m_fWidthDim;
			fYHDim = m_fHeightDim;
			fZDDim = m_fLengthDim;
			iID = i + (m_iWidthSegs+1)*j;
			if(iSurf<3) iID += (m_iHeightSegs+1)*(m_iWidthSegs+1)*m_iLengthSegs;
			break;
		default:
			assert(iSurf>=0 && iSurf<=2/*, "Found no matching surface\n"*/);
	}

	int Index = 0;
	int * pIDs = m_iPointID;//(int *) (((unsigned int) m_iPointID) | 0x30000000);
	int iPntID = pIDs[Index];
	while(iPntID!=-1 && iPntID!=iID)
	{
		++Index;
		iPntID = pIDs[Index];
		//assert(Index<m_iNumOfUniqueVerts/*, "ID test failed\n"*/);
	}
	

	// Make new point
	if(m_iPointID[Index]!=iID)	
	{
		//m_iPointID[Index]=iID;

		float fTx, fTy, fTz;
		float fX, fY, fZ;

		fTx  = (float(i)/iXWSeg)-0.5f;
		fTy  = (float(j)/iYHSeg)-0.5f;
		fTz  = (iSurf>=3 /* is Counter Surf*/) ? -0.5f : 0.5f;

		fX = fTx * fXWDim;
		fY = fTy * fYHDim;
		fZ = fTz * fZDDim;

		//float fDelZ = (fRnd()*10)-5.0f;
		//float fDelZ = (fRnd()*50)-25;

		m_pUniVerts[Index*4 + 0] = fX;
		m_pUniVerts[Index*4 + 1] = fY;
		m_pUniVerts[Index*4 + 2] = fZ;// + fDelZ;
		m_pUniVerts[Index*4 + 3] = 1;

		m_pUniCols[Index*4 + 0] = 127;//(int) (fRnd()*127);
		m_pUniCols[Index*4 + 1] = 127;//(int) (fRnd()*127);
		m_pUniCols[Index*4 + 2] = 127;//(int) (fRnd()*127);
		m_pUniCols[Index*4 + 3] = 0x3f;

		// Hack!
		//m_pUniCols[Index*4 + 1] = m_pUniCols[Index*4 + 0];
		//m_pUniCols[Index*4 + 2] = m_pUniCols[Index*4 + 1];
		//m_pUniCols[Index*4 + 3] = m_pUniCols[Index*4 + 2];

		switch(iSurf%3)
		{
			float fT;

			case 0:
				fT = m_pUniVerts[Index*4 + 0];
				m_pUniVerts[Index*4 + 0] = m_pUniVerts[Index*4 + 2];
				m_pUniVerts[Index*4 + 2] = fT;
				break;
			case 1:
				fT = m_pUniVerts[Index*4 + 1];
				m_pUniVerts[Index*4 + 1] = m_pUniVerts[Index*4 + 2];
				m_pUniVerts[Index*4 + 2] = fT;
				break;
		}
	}

	pVert[0] = m_pUniVerts[Index*4 + 0];
	pVert[1] = m_pUniVerts[Index*4 + 1];
	pVert[2] = m_pUniVerts[Index*4 + 2];
	pVert[3] = m_pUniVerts[Index*4 + 3];
	pCol[0]  = m_pUniCols[Index*4 + 0];
	pCol[1]  = m_pUniCols[Index*4 + 1];
	pCol[2]  = m_pUniCols[Index*4 + 2];
	pCol[3]  = m_pUniCols[Index*4 + 3];
}


CBoxMesh::CBoxMesh()
{
	m_pUniCols = NULL;
	m_pUniVerts = NULL;
	m_iPointID = NULL;


	m_iNumOfUniqueVerts = 0;
}


void CBoxMesh::CleanUp()
{
	IMesh::CleanUp();

	if(m_pUniCols!=NULL) delete [] m_pUniCols;
	if(m_pUniVerts!=NULL) delete [] m_pUniVerts;
	if(m_iPointID!=NULL) delete [] m_iPointID;

	m_iNumOfUniqueVerts = 0;
}

void CBoxMesh::CleanStrips(SRenderGadget * pStrips, int iTotStrips)
{
	if(pStrips != NULL)
	{
		for(int i=0; i<iTotStrips; i++)
		{
			if(pStrips[i].m_pVerts != NULL)
				delete [] pStrips[i].m_pVerts;

			if(pStrips[i].m_pCols != NULL)
				delete [] pStrips[i].m_pCols;

			if(pStrips[i].m_pTexCrds != NULL)
				delete [] pStrips[i].m_pTexCrds;
		}

		delete pStrips;
	}
}
