/*
    Copyright (C) 2002 Terratron Technologies Inc.  All Rights Reserved.
    Author: Steven Osman
 
    This file is part of sps2.

    For more information visit:
    http://window.terratron.com/~sosman/ps2linux/
    http://www.terratron.com/
*/
#ifndef __SPS2_TAGS_H
#define __SPS2_TAGS_H

#include "sps2registers.h"
  
#ifdef __cplusplus
extern "C" {
#endif

#define SPS2_SET_GIF_REG(gifTag, iRegNo, uiValue) \
  {((sps2GIFTag_t *) &gifTag)->ai64[1] |= ((uiValue & 0xf) << (iRegNo << 2));}

  typedef enum GIF_REG {
    GIF_REG_PRIM      =0x00,
    GIF_REG_RGBAQ     =0x01,
    GIF_REG_ST        =0x02,
    GIF_REG_UV        =0x03,
    GIF_REG_XYZF2     =0x04,
    GIF_REG_XYZ2      =0x05,
    GIF_REG_TEX0_1    =0x06,
    GIF_REG_TEX0_2    =0x07,
    GIF_REG_CLAMP_1   =0x08,
    GIF_REG_CLAMP_2   =0x09,
    GIF_REG_FOG       =0x0a,
    GIF_REG_XYZF3     =0x0c,
    GIF_REG_XYZ3      =0x0d,
    GIF_REG_A_D       =0x0e,
    GIF_REG_NOP       =0x0f
  } GIF_REG_t;

  typedef enum GIF_FLG {
    GIF_FLG_PACKED    =0x0,
    GIF_FLG_REGLIST   =0x1,
    GIF_FLG_IMAGE     =0x2,
  } GIF_FLG_t;


  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int NLOOP : 15;
      unsigned int EOP   : 1;
      unsigned int _PAD1 : 16;
      unsigned int _PAD2 : 14;
      unsigned int PRE   : 1;
      unsigned int PRIM  : 11;
      GIF_FLG_t    FLG   : 2;
      unsigned int NREG  : 4;
      sps2uint64 REGS  : 64;
    } s;

    sps2uint64 ai64[2];
  } sps2GIFTag_t;

  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int PRIM: 11;
      unsigned int _PAD1: 21;
      unsigned int _PAD2: 32;
      unsigned int _PAD3: 32;
      unsigned int _PAD4: 32;
    } s;
  } sps2GIFPackedPRIM_t;

  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int R: 8;
      unsigned int _PAD1: 24;
      unsigned int G: 8;
      unsigned int _PAD2: 24;
      unsigned int B: 8;
      unsigned int _PAD3: 24;
      unsigned int A: 8;
      unsigned int _PAD4: 24;
    } s;
  } sps2GIFPackedRGBAQ_t;

  typedef union {
    sps2uint128 i128;

    struct {
      float S;
      float T;
      float Q;
      unsigned int _PAD1: 32;
    } s;
  } sps2GIFPackedST_t;

  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int U: 14;
      unsigned int _PAD1: 18;
      unsigned int V: 14;
      unsigned int _PAD2: 18;
      unsigned int _PAD3: 32;
      unsigned int _PAD4: 32;
    } s;
  } sps2GIFPackedUV_t;

  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int X: 16;
      unsigned int _PAD1: 16;
      unsigned int Y: 16;
      unsigned int _PAD2: 16;
      unsigned int _PAD3: 4;
      unsigned int Z: 24;
      unsigned int _PAD4: 4;
      unsigned int _PAD5: 4;
      unsigned int F: 8;
      unsigned int _PAD6: 3;
      unsigned int ADC: 1;
      unsigned int _PAD7: 16;
    } s;
  } sps2GIFPackedXYZF2_t;

  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int X: 16;
      unsigned int _PAD1: 16;
      unsigned int Y: 16;
      unsigned int _PAD2: 16;
      unsigned int Z: 32;
      unsigned int _PAD3: 15;
      unsigned int ADC: 1;
      unsigned int _PAD4: 16;
    } s;
  } sps2GIFPackedXYZ2_t;
  
  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int _PAD1: 32;
      unsigned int _PAD2: 32;
      unsigned int _PAD3: 32;
      unsigned int _PAD4: 4;
      unsigned int F: 8;
      unsigned int _PAD5: 20;
    } s;
  } sps2GIFPackedFOG_t;
  
  typedef union {
    sps2uint128 i128;

    struct {
      sps2uint64 DATA: 64;
      unsigned int ADDR: 8;
      unsigned int _PAD1: 24;
      unsigned int _PAD2: 32;
    } s;
  } sps2GIFPackedA_D_t;

  typedef union {
    sps2uint128 i128;

    struct {
      unsigned int _PAD1: 32;
      unsigned int _PAD2: 32;
      unsigned int _PAD3: 32;
      unsigned int _PAD4: 32;
    } s;
  } sps2GIFPackedNOP_t;

  typedef union {
    sps2uint128          i128;
    sps2GIFPackedPRIM_t  PRIM;
    sps2GIFPackedRGBAQ_t RGBAQ;
    sps2GIFPackedST_t    ST;
    sps2GIFPackedUV_t    UV;
    sps2GIFPackedXYZF2_t XYZF2;
    sps2GIFPackedXYZ2_t  XYZ2;
    sps2GIFPackedFOG_t   FOG;
    sps2GIFPackedA_D_t   A_D;
    sps2GIFPackedNOP_t   NOP;
  } sps2GIFPackedRegister_t;

  // Used by CHCR and the DMA tag
  typedef enum DMA_ID {
    DMA_ID_SOURCE_REFE=0,
    DMA_ID_SOURCE_CNT =1,
    DMA_ID_SOURCE_NEXT=2,
    DMA_ID_SOURCE_REF =3,
    DMA_ID_SOURCE_REFS=4,
    DMA_ID_SOURCE_CALL=5,
    DMA_ID_SOURCE_RET =6,
    DMA_ID_SOURCE_END =7, 
    DMA_ID_DEST_CNTS  =0,
    DMA_ID_DEST_CNT   =1,
    DMA_ID_DEST_END   =7,
  } DMA_ID_t;

  typedef union {
    sps2uint64 i64;
    
    struct {
      unsigned int QWC    : 16;
      unsigned int _PAD1  : 10;
      unsigned int PCE    :  2;
      DMA_ID_t     ID     :  3;
      unsigned int IRQ    :  1;
      unsigned int ADDR   : 31;
      unsigned int SPR    :  1;
      unsigned int _PAD2  : 32;
      unsigned int _PAD3  : 32;
    } s;
  } sps2DMATag_t;

#ifdef __cplusplus
}
#endif
#endif
